package com.sun.jmx.mbeanserver;

import java.lang.reflect.Constructor;
import javax.management.MBeanConstructorInfo;
import javax.management.NotificationBroadcaster;
import javax.management.ImmutableDescriptor;
import java.util.WeakHashMap;
import javax.management.MBeanNotificationInfo;
import java.lang.reflect.Array;
import javax.management.InvalidAttributeValueException;
import javax.management.ReflectionException;
import javax.management.MBeanInfo;
import java.lang.ref.WeakReference;
import java.util.Arrays;
import sun.reflect.misc.ReflectUtil;
import java.util.List;
import javax.management.Descriptor;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Type;
import java.lang.reflect.Method;
import javax.management.NotCompliantMBeanException;

abstract class MBeanIntrospector<M>
{
    MBeanIntrospector() {
    }
    
    abstract MBeanIntrospector.PerInterfaceMap<M> getPerInterfaceMap();
    
    abstract MBeanIntrospector.MBeanInfoMap getMBeanInfoMap();
    
    abstract MBeanAnalyzer<M> getAnalyzer(Class<?> p0) throws NotCompliantMBeanException;
    
    abstract boolean isMXBean();
    
    abstract M mFrom(Method p0);
    
    abstract String getName(M p0);
    
    abstract Type getGenericReturnType(M p0);
    
    abstract Type[] getGenericParameterTypes(M p0);
    
    abstract String[] getSignature(M p0);
    
    abstract void checkMethod(M p0);
    
    abstract Object invokeM2(M p0, Object p1, Object[] p2, Object p3) throws InvocationTargetException, IllegalAccessException, MBeanException;
    
    abstract boolean validParameter(M p0, Object p1, int p2, Object p3);
    
    abstract MBeanAttributeInfo getMBeanAttributeInfo(String p0, M p1, M p2);
    
    abstract MBeanOperationInfo getMBeanOperationInfo(String p0, M p1);
    
    abstract Descriptor getBasicMBeanDescriptor();
    
    abstract Descriptor getMBeanDescriptor(Class<?> p0);
    
    final List<Method> getMethods(Class<?> clazz) {
        ReflectUtil.checkPackageAccess(clazz);
        return Arrays.asList(clazz.getMethods());
    }
    
    final PerInterface<M> getPerInterface(Class<?> clazz) throws NotCompliantMBeanException {
        MBeanIntrospector.PerInterfaceMap<M> perInterfaceMap = this.getPerInterfaceMap();
        synchronized (perInterfaceMap) {
            WeakReference<Object> weakReference = (WeakReference)perInterfaceMap.get(clazz);
            PerInterface referent = (weakReference == null) ? null : ((PerInterface)weakReference.get());
            if (referent == null) {
                try {
                    MBeanAnalyzer<M> analyzer = this.getAnalyzer(clazz);
                    referent = new PerInterface(clazz, this, analyzer, this.makeInterfaceMBeanInfo(clazz, analyzer));
                    perInterfaceMap.put(clazz, new WeakReference(referent));
                }
                catch (Exception ex) {
                    throw Introspector.throwException(clazz, ex);
                }
            }
            return referent;
        }
    }
    
    private MBeanInfo makeInterfaceMBeanInfo(Class<?> clazz, MBeanAnalyzer<M> mBeanAnalyzer) {
        MBeanIntrospector.MBeanInfoMaker mBeanInfoMaker = new MBeanIntrospector.MBeanInfoMaker();
        mBeanAnalyzer.visit(mBeanInfoMaker);
        return mBeanInfoMaker.makeMBeanInfo(clazz, "Information on the management interface of the MBean");
    }
    
    final boolean consistent(M m, M i) {
        return m == null || i == null || this.getGenericReturnType(m).equals(this.getGenericParameterTypes(i)[0]);
    }
    
    final Object invokeM(M m, Object o, Object[] array, Object o2) throws MBeanException, ReflectionException {
        try {
            return this.invokeM2(m, o, array, o2);
        }
        catch (InvocationTargetException cause) {
            unwrapInvocationTargetException(cause);
            throw new RuntimeException(cause);
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, e.toString());
        }
    }
    
    final void invokeSetter(String s, M m, Object o, Object o2, Object o3) throws MBeanException, ReflectionException, InvalidAttributeValueException {
        try {
            this.invokeM2(m, o, new Object[] { o2 }, o3);
        }
        catch (IllegalAccessException e) {
            throw new ReflectionException(e, e.toString());
        }
        catch (RuntimeException ex) {
            this.maybeInvalidParameter(s, m, o2, o3);
            throw ex;
        }
        catch (InvocationTargetException ex2) {
            this.maybeInvalidParameter(s, m, o2, o3);
            unwrapInvocationTargetException(ex2);
        }
    }
    
    private void maybeInvalidParameter(String str, M m, Object obj, Object o) throws InvalidAttributeValueException {
        if (!this.validParameter(m, obj, 0, o)) {
            throw new InvalidAttributeValueException("Invalid value for attribute " + str + ": " + obj);
        }
        return;
    }
    
    static boolean isValidParameter(Method method, Object o, int n) {
        Class<Object> componentType = method.getParameterTypes()[n];
        try {
            Array.set(Array.newInstance(componentType, 1), 0, o);
            return true;
        }
        catch (IllegalArgumentException ex) {
            return false;
        }
    }
    
    private static void unwrapInvocationTargetException(InvocationTargetException ex) throws MBeanException {
        Throwable cause = ex.getCause();
        if ((cause instanceof RuntimeException)) {
            throw (RuntimeException)cause;
        }
        if ((cause instanceof Error)) {
            throw (Error)cause;
        }
        throw new MBeanException((Exception)cause, (cause == null) ? null : cause.toString());
    }
    
    final MBeanInfo getMBeanInfo(Object o, PerInterface<M> perInterface) {
        MBeanInfo classMBeanInfo = this.getClassMBeanInfo(o.getClass(), perInterface);
        MBeanNotificationInfo[] notifications = findNotifications(o);
        if (notifications == null || notifications.length == 0) {
            return classMBeanInfo;
        }
        return new MBeanInfo(classMBeanInfo.getClassName(), classMBeanInfo.getDescription(), classMBeanInfo.getAttributes(), classMBeanInfo.getConstructors(), classMBeanInfo.getOperations(), notifications, classMBeanInfo.getDescriptor());
    }
    
    final MBeanInfo getClassMBeanInfo(Class<?> clazz, PerInterface<M> perInterface) {
        MBeanIntrospector.MBeanInfoMap mBeanInfoMap = this.getMBeanInfoMap();
        synchronized (mBeanInfoMap) {
            WeakHashMap<Class, MBeanInfo> value = (WeakHashMap)mBeanInfoMap.get(clazz);
            if (value == null) {
                value = new WeakHashMap();
                mBeanInfoMap.put(clazz, value);
            }
            Class<Object> mBeanInterface = perInterface.getMBeanInterface();
            MBeanInfo value2 = (Object)value.get(mBeanInterface);
            if (value2 == null) {
                MBeanInfo mBeanInfo = perInterface.getMBeanInfo();
                value2 = new MBeanInfo(clazz.getName(), mBeanInfo.getDescription(), mBeanInfo.getAttributes(), findConstructors(clazz), mBeanInfo.getOperations(), (MBeanNotificationInfo[])null, ImmutableDescriptor.union(new Descriptor[] { mBeanInfo.getDescriptor(), this.getMBeanDescriptor(clazz) }));
                value.put(mBeanInterface, value2);
                return value2;
            }
            return value2;
        }
    }
    
    static MBeanNotificationInfo[] findNotifications(Object o) {
        if (!(o instanceof NotificationBroadcaster)) {
            return null;
        }
        MBeanNotificationInfo[] notificationInfo = ((NotificationBroadcaster)o).getNotificationInfo();
        if (notificationInfo == null) {
            return null;
        }
        MBeanNotificationInfo[] array = new MBeanNotificationInfo[notificationInfo.length];
        MBeanNotificationInfo mBeanNotificationInfo;
        for (int i = 0; i < notificationInfo.length; array[i] = mBeanNotificationInfo, ++i) {
            mBeanNotificationInfo = notificationInfo[i];
            if (mBeanNotificationInfo.getClass() != MBeanNotificationInfo.class) {
                mBeanNotificationInfo = (MBeanNotificationInfo)mBeanNotificationInfo.clone();
            }
        }
        return array;
    }
    
    private static MBeanConstructorInfo[] findConstructors(Class<?> clazz) {
        Constructor[] constructors = clazz.getConstructors();
        MBeanConstructorInfo[] array = new MBeanConstructorInfo[constructors.length];
        for (int i = 0; i < constructors.length; ++i) {
            array[i] = new MBeanConstructorInfo("Public constructor of the MBean", constructors[i]);
        }
        return array;
    }
    
    static final class PerInterfaceMap<M> extends WeakHashMap<Class<?>, WeakReference<PerInterface<M>>>
    {
    }
    
    private class MBeanInfoMaker implements MBeanAnalyzer.MBeanVisitor<M>
    {
        private final List<MBeanAttributeInfo> attrs;
        private final List<MBeanOperationInfo> ops;
        
        private MBeanInfoMaker() {
            this.attrs = Util.newList();
            this.ops = Util.newList();
        }
        
        @Override
        public void visitAttribute(String s, M m, M i) {
            this.attrs.add(MBeanIntrospector.this.getMBeanAttributeInfo(s, m, i));
        }
        
        @Override
        public void visitOperation(String s, M m) {
            this.ops.add(MBeanIntrospector.this.getMBeanOperationInfo(s, m));
        }
        
        MBeanInfo makeMBeanInfo(Class<?> clazz, String description) {
            return new MBeanInfo(clazz.getName(), description, (MBeanAttributeInfo[])this.attrs.toArray(new MBeanAttributeInfo[0]), null, (MBeanOperationInfo[])this.ops.toArray(new MBeanOperationInfo[0]), null, DescriptorCache.getInstance().union(new Descriptor[] { new ImmutableDescriptor(new String[] { "interfaceClassName=" + clazz.getName() }), MBeanIntrospector.this.getBasicMBeanDescriptor(), Introspector.descriptorForElement(clazz) }));
        }
    }
    
    static class MBeanInfoMap extends WeakHashMap<Class<?>, WeakHashMap<Class<?>, MBeanInfo>>
    {
    }
}
